@@ -31,9 +31,22 @@ import re
 import sys
 from optparse import SUPPRESS_HELP, OptionGroup, OptionParser
 
-from nuitka.utils import Utils
+from nuitka.PythonFlavors import (
+    isAnacondaPython,
+    isApplePython,
+    isDebianPackagePython,
+    isNuitkaPython,
+    isPyenvPython,
+    isWinPython,
+)
 from nuitka.utils.FileOperations import getFileContentByLine
-from nuitka.Version import getNuitkaVersion
+from nuitka.utils.Utils import (
+    getArchitecture,
+    getLinuxDistribution,
+    getOS,
+    isLinux,
+)
+from nuitka.Version import getCommercialVersion, getNuitkaVersion
 
 # Indicator if we were called as "nuitka-run" in which case we assume some
 # other defaults and work a bit different with parameters.
@@ -44,17 +57,44 @@ if not is_nuitka_run:
 else:
     usage = "usage: %prog [options] main_module.py"
 
+
+def _getPythonFlavor():
+    # return driven, pylint: disable=too-many-return-statements
+
+    if isNuitkaPython():
+        return "Nuitka Python"
+    elif isAnacondaPython():
+        return "Anaconda Python"
+    elif isWinPython():
+        return "WinPython"
+    elif isDebianPackagePython():
+        return "Debian Python"
+    elif isApplePython():
+        return "Apple Python"
+    elif isPyenvPython():
+        return "pyenv"
+    else:
+        return "Unknown"
+
+
+def _getVersionInformationValues():
+    # TODO: Might be nice if we could delay version information computation
+    # until it's actually used.
+    yield getNuitkaVersion()
+    yield "Commercial: %s" % getCommercialVersion()
+    yield "Python: %s" % sys.version.split("\n", 1)[0]
+    yield "Flavor: %s" % _getPythonFlavor()
+    yield "Executable: %s" % sys.executable
+    yield "OS: %s" % getOS()
+    yield "Arch: %s" % getArchitecture()
+
+    if isLinux():
+        yield "Distribution: %s %s" % getLinuxDistribution()
+
+
 parser = OptionParser(
     usage=usage,
-    version="\n".join(
-        (
-            getNuitkaVersion(),
-            "Python: " + sys.version.split("\n")[0],
-            "Executable: " + sys.executable,
-            "OS: " + Utils.getOS(),
-            "Arch: " + Utils.getArchitecture(),
-        )
-    ),
+    version="\n".join(_getVersionInformationValues()),
 )
 
 parser.add_option(
@@ -117,7 +157,7 @@ if os.name == "nt":
         help="""\
 Architecture of Python to use. One of "x86" or "x86_64".
 Defaults to what you run Nuitka with (currently "%s")."""
-        % (Utils.getArchitecture()),
+        % (getArchitecture()),
     )
 
 parser.add_option(
@@ -134,19 +174,22 @@ parser.add_option(
     "--python-flag",
     action="append",
     dest="python_flags",
+    metavar="FLAG",
     default=[],
     help="""\
-Python flags to use. Default uses what you are using to run Nuitka, this
+Python flags to use. Default is what you are using to run Nuitka, this
 enforces a specific mode. These are options that also exist to standard
-Python executable. Currently supported: "-S" (alias "nosite"),
+Python executable. Currently supported: "-S" (alias "no_site"),
 "static_hashes" (do not use hash randomization), "no_warnings" (do not
-give Python runtime warnings), "-O" (alias "noasserts"). Default empty.""",
+give Python runtime warnings), "-O" (alias "no_asserts"), "no_docstrings"
+(do not use docstrings). Default empty.""",
 )
 
 parser.add_option(
     "--python-for-scons",
     action="store",
     dest="python_scons",
+    metavar="PATH",
     default=None,
     help="""\
 If using Python3.3 or Python3.4, provide the path of a Python binary to use
@@ -178,7 +221,8 @@ parser.add_option(
     dest="assume_yes_for_downloads",
     default=False,
     help="""\
-Allow Nuitka to download code if necessary, e.g. dependency walker on Windows.""",
+Allow Nuitka to download external code if necessary, e.g. dependency
+walker, ccache, and even gcc on Windows.""",
 )
 
 
@@ -193,10 +237,11 @@ include_group.add_option(
     metavar="PACKAGE",
     default=[],
     help="""\
-Include a whole package. Give as a Python namespace, e.g. ``some_package.sub_package``
+Include a whole package. Give as a Python namespace, e.g. "some_package.sub_package"
 and Nuitka will then find it and include it and all the modules found below that
 disk location in the binary or extension module it creates, and make it available
-for import by the code. Default empty.""",
+for import by the code. To avoid unwanted sub packages, e.g. tests you can e.g. do
+this "--nofollow-import-to=*.tests". Default empty.""",
 )
 
 include_group.add_option(
@@ -206,7 +251,7 @@ include_group.add_option(
     metavar="MODULE",
     default=[],
     help="""\
-Include a single module. Give as a Python namespace, e.g. ``some_package.some_module``
+Include a single module. Give as a Python namespace, e.g. "some_package.some_module"
 and Nuitka will then find it and include it in the binary or extension module
 it creates, and make it available for import by the code. Default empty.""",
 )
@@ -214,19 +259,19 @@ it creates, and make it available for import by the code. Default empty.""",
 include_group.add_option(
     "--include-plugin-directory",
     action="append",
-    dest="recurse_extra",
+    dest="include_extra",
     metavar="MODULE/PACKAGE",
     default=[],
     help="""\
 Include the content of that directory, no matter if it's used by the given main
-program in a visible form. Overrides all other recursion options. Can be given
+program in a visible form. Overrides all other inclusion options. Can be given
 multiple times. Default empty.""",
 )
 
 include_group.add_option(
     "--include-plugin-files",
     action="append",
-    dest="recurse_extra_files",
+    dest="include_extra_files",
     metavar="PATTERN",
     default=[],
     help="""\
@@ -263,7 +308,7 @@ follow_group = OptionGroup(parser, "Control the following into imported modules"
 follow_group.add_option(
     "--follow-stdlib",
     action="store_true",
-    dest="recurse_stdlib",
+    dest="follow_stdlib",
     default=False,
     help="""\
 Also descend into imported modules from standard library. This will increase
@@ -277,7 +322,7 @@ follow_group.add_option(
     default=False,
     help="""\
 When --nofollow-imports is used, do not descend into any imported modules at all,
-overrides all other recursion options. Defaults to off.""",
+overrides all other inclusion options. Defaults to off.""",
 )
 
 follow_group.add_option(
@@ -322,7 +367,7 @@ data_group.add_option(
     "--include-package-data",
     action="append",
     dest="package_data",
-    metavar="PACKAGE_DATA",
+    metavar="PACKAGE",
     default=[],
     help="""\
 Include data files of the given package name. Can use patterns. By default
@@ -335,7 +380,7 @@ data_group.add_option(
     "--include-data-file",
     action="append",
     dest="data_files",
-    metavar="DATA_FILES",
+    metavar="DESC",
     default=[],
     help="""\
 Include data files by filenames in the distribution. There are many
@@ -351,7 +396,7 @@ data_group.add_option(
     "--include-data-dir",
     action="append",
     dest="data_dirs",
-    metavar="DATA_DIRS",
+    metavar="DIRECTORY",
     default=[],
     help="""\
 Include data files from complete directory in the distribution. This is
@@ -361,6 +406,25 @@ for plain copy, of the whole directory. All files are copied, if you want to
 exclude files you need to remove them beforehand. Default empty.""",
 )
 
+data_files_tags = [("inhibit", "do not include the file")]
+
+# TODO: Expose this when finished, pylint: disable=using-constant-test
+if False:
+    data_group.add_option(
+        "--data-file-tags",
+        action="append",
+        dest="data_tags",
+        metavar="DATA_TAGS",
+        default=[],
+        help="""\
+    For included data files, special handlings can be chosen. With the
+    commercial plugins, e.g. files can be included directly in the
+    binary. The list is completed by some plugins. With the current
+    list of plugins, these are available: %s.
+    The default is empty."""
+        % ",".join("'%s' (%s)" % d for d in data_files_tags),
+    )
+
 
 parser.add_option_group(data_group)
 
@@ -435,6 +499,7 @@ codegen_group.add_option(
     "--file-reference-choice",
     action="store",
     dest="file_reference_mode",
+    metavar="MODE",
     choices=("original", "runtime", "frozen"),
     default=None,
     help="""\
@@ -466,7 +531,7 @@ include path information that needs to exist though. Defaults to '%s' on this
 platform.
 """
     % "<program_name>"
-    + (".exe" if Utils.getOS() == "Windows" else ".bin"),
+    + (".exe" if getOS() == "Windows" else ".bin"),
 )
 
 output_group.add_option(
@@ -584,6 +649,7 @@ debug_group.add_option(
     "--experimental",
     action="append",
     dest="experimental",
+    metavar="FLAG",
     default=[],
     help="""\
 Use features declared as 'experimental'. May have no effect if no experimental
@@ -599,6 +665,17 @@ debug_group.add_option(
     help=SUPPRESS_HELP,
 )
 
+debug_group.add_option(
+    "--low-memory",
+    action="store_true",
+    dest="low_memory",
+    default=False,
+    help="""\
+Attempt to use less memory, by forking less C compilation jobs and using
+options that use less memory. For use on embedded machines. Use this in
+case of out of memory problems. Defaults to off.""",
+)
+
 if os.name == "nt":
     debug_group.add_option(
         "--disable-dll-dependency-cache",
@@ -650,28 +727,28 @@ Enforce the use of clang. On Windows this requires a working Visual
 Studio version to piggy back on. Defaults to off.""",
 )
 
-if os.name == "nt":
-    c_compiler_group.add_option(
-        "--mingw64",
-        action="store_true",
-        dest="mingw64",
-        default=False,
-        help="""\
+c_compiler_group.add_option(
+    "--mingw64",
+    action="store_true",
+    dest="mingw64",
+    default=False,
+    help="""\
 Enforce the use of MinGW64 on Windows. Defaults to off.""",
-    )
+)
 
-    c_compiler_group.add_option(
-        "--msvc",
-        action="store",
-        dest="msvc",
-        default=None,
-        help="""\
+c_compiler_group.add_option(
+    "--msvc",
+    action="store",
+    dest="msvc_version",
+    default=None,
+    help="""\
 Enforce the use of specific MSVC version on Windows. Allowed values
-are e.g. 14.0, specify an illegal value for a list of installed compilers,
-beware that only latest MSVC is really supported.
+are e.g. "14.2" (MSVC 2019), specify an illegal value for a list of
+installed compilers, or use "latest". Notice that only latest
+MSVC is really supported, and you can use "latest" to enforce that.
 
-Defaults to the most recent version.""",
-    )
+Defaults to MSVC on Windows being used if installed, otherwise MinGW64.""",
+)
 
 c_compiler_group.add_option(
     "-j",
@@ -679,7 +756,7 @@ c_compiler_group.add_option(
     action="store",
     dest="jobs",
     metavar="N",
-    default=Utils.getCoreCount(),
+    default=None,
     help="""\
 Specify the allowed number of parallel C compiler jobs. Defaults to the
 system CPU count.""",
@@ -687,27 +764,102 @@ system CPU count.""",
 
 c_compiler_group.add_option(
     "--lto",
-    action="store_true",
+    action="store",
     dest="lto",
-    default=False,
+    metavar="choice",
+    default="auto",
+    choices=("yes", "no", "auto"),
     help="""\
-Use link time optimizations if available and usable (MSVC, gcc >=4.6, clang).
-Defaults to off.""",
+Use link time optimizations (MSVC, gcc, clang). Allowed values are
+"yes", "no", and "auto" (when it's known to work). Defaults to
+"auto".""",
 )
 
 c_compiler_group.add_option(
     "--static-libpython",
     action="store",
     dest="static_libpython",
+    metavar="choice",
     default="auto",
     choices=("yes", "no", "auto"),
     help="""\
-Use static link library of Python if available. Defaults to auto, i.e. enabled for where
-we know it's working.""",
+Use static link library of Python. Allowed values are "yes", "no",
+and "auto" (when it's known to work). Defaults to "auto".""",
+)
+
+c_compiler_group.add_option(
+    "--disable-ccache",
+    action="store_true",
+    dest="no_ccache",
+    default=False,
+    help="""\
+Do not attempt to use ccache (gcc, clang, etc.) or clcache (MSVC, clangcl).""",
 )
 
 parser.add_option_group(c_compiler_group)
 
+pgo_group = OptionGroup(parser, "PGO compilation choices")
+
+pgo_group.add_option(
+    "--pgo",
+    action="store_true",
+    dest="is_c_pgo",
+    default=False,
+    help="""\
+Enables C level profile guided optimization (PGO), by executing a dedicated build first
+for a profiling run, and then using the result to feedback into the C compilation.
+Note: This is experimental and not working with standalone modes of Nuitka yet.
+Defaults to off.""",
+)
+
+pgo_group.add_option(
+    "--pgo-python",
+    action="store_true",
+    dest="is_python_pgo",
+    default=False,
+    help=SUPPRESS_HELP,
+)
+
+pgo_group.add_option(
+    "--pgo-python-input",
+    action="store",
+    dest="python_pgo_input",
+    default=None,
+    help=SUPPRESS_HELP,
+)
+
+pgo_group.add_option(
+    "--pgo-python-policy-unused-module",
+    action="store",
+    dest="python_pgo_policy_unused_module",
+    choices=("include", "exclude", "bytecode"),
+    default="include",
+    help=SUPPRESS_HELP,
+)
+
+pgo_group.add_option(
+    "--pgo-args",
+    action="store",
+    dest="pgo_args",
+    default="",
+    help="""\
+Arguments to be passed in case of profile guided optimization. These are passed to the special
+built executable during the PGO profiling run. Default empty.""",
+)
+
+pgo_group.add_option(
+    "--pgo-executable",
+    action="store",
+    dest="pgo_executable",
+    default=None,
+    help="""\
+Command to execute when collecting profile information. Use this only, if you need to
+launch it through a script that prepares it to run. Default use created program.""",
+)
+
+
+parser.add_option_group(pgo_group)
+
 tracing_group = OptionGroup(parser, "Tracing features")
 
 tracing_group.add_option(
@@ -740,7 +892,7 @@ Defaults to off.""",
 )
 
 tracing_group.add_option(
-    "--no-progress",
+    "--no-progressbar",
     action="store_false",
     dest="progress_bar",
     default=True,
@@ -773,6 +925,7 @@ tracing_group.add_option(
     "--show-modules-output",
     action="store",
     dest="show_inclusion_output",
+    metavar="PATH",
     default=None,
     help="""\
 Where to output --show-modules, should be a filename. Default is standard output.""",
@@ -792,6 +945,7 @@ tracing_group.add_option(
     "--verbose-output",
     action="store",
     dest="verbose_output",
+    metavar="PATH",
     default=None,
     help="""\
 Where to output --verbose, should be a filename. Default is standard output.""",
@@ -812,7 +966,7 @@ windows_group.add_option(
 windows_group.add_option(
     "--windows-disable-console",
     action="store_true",
-    dest="win_disable_console",
+    dest="disable_console",
     default=False,
     help="""\
 When compiling for Windows, disable the console window. Defaults to off.""",
@@ -840,6 +994,15 @@ windows_group.add_option(
     help="Copy executable icons from this existing executable (Windows only).",
 )
 
+windows_group.add_option(
+    "--onefile-windows-splash-screen-image",
+    action="store",
+    dest="splash_screen_image",
+    default=None,
+    help="""\
+When compiling for Windows and onefile, show this while loading the application. Defaults to off.""",
+)
+
 windows_group.add_option(
     "--windows-uac-admin",
     action="store_true",
@@ -893,7 +1056,7 @@ windows_group.add_option(
     default=None,
     help="""\
 File version to use in Windows Version information. Must be a sequence of
-up to 4 numbers, nothing else allowed.
+up to 4 numbers, e.g. 1.0.0.0, only this format is allowed.
 One of file or product version is required, when a version resource needs to be
 added, e.g. to specify product name, or company name. Defaults to unused.""",
 )
@@ -906,7 +1069,7 @@ windows_group.add_option(
     default=None,
     help="""\
 Product version to use in Windows Version information. Must be a sequence of
-up to 4 numbers, nothing else allowed.
+up to 4 numbers, e.g. 1.0.0.0, only this format is allowed.
 One of file or product version is required, when a version resource needs to be
 added, e.g. to specify product name, or company name. Defaults to unused.""",
 )
@@ -971,6 +1134,75 @@ to not active, use e.g. '%PROGRAM%.err.txt', i.e. file near your program.""",
 
 parser.add_option_group(windows_group)
 
+macos_group = OptionGroup(parser, "macOS specific controls")
+
+macos_group.add_option(
+    "--macos-onefile-icon",
+    action="append",
+    dest="icon_path",
+    metavar="ICON_PATH",
+    default=[],
+    help="Add executable icon for binary to use. Can be given only one time. Defaults to Python icon if available.",
+)
+
+macos_group.add_option(
+    "--macos-disable-console",
+    action="store_true",
+    dest="disable_console",
+    default=False,
+    help="""\
+When compiling for macOS, disable the console window and create a GUI
+application. Defaults to off.""",
+)
+
+macos_group.add_option(
+    "--macos-create-app-bundle",
+    action="store_true",
+    dest="macos_create_bundle",
+    default=False,
+    help="""\
+When compiling for macOS, create a bundle rather than a plain binary
+application. Currently experimental and incomplete. Currently this
+is the only way to unlock disabling of console.Defaults to off.""",
+)
+
+macos_group.add_option(
+    "--macos-signed-app-name",
+    action="store",
+    dest="macos_signed_app_name",
+    metavar="MACOS_SIGNED_APP_NAME",
+    default=None,
+    help="""\
+Name of the application to use for macOS signing. Follow com.yourcompany.appname naming
+results for best results, as these have to be globally unique, and will grant protected
+API accesses.""",
+)
+
+macos_group.add_option(
+    "--macos-app-name",
+    action="store",
+    dest="macos_app_name",
+    metavar="MACOS_APP_NAME",
+    default=None,
+    help="""\
+Name of the product to use in macOS bundle information. Defaults to base
+filename of the binary.""",
+)
+
+macos_group.add_option(
+    "--macos-app-version",
+    action="store",
+    dest="macos_app_version",
+    metavar="MACOS_APP_VERSION",
+    default=None,
+    help="""\
+Product version to use in macOS bundle information. Defaults to 1.0 if
+not given.""",
+)
+
+
+parser.add_option_group(macos_group)
+
 linux_group = OptionGroup(parser, "Linux specific controls")
 
 linux_group.add_option(
@@ -987,10 +1219,11 @@ parser.add_option_group(linux_group)
 plugin_group = OptionGroup(parser, "Plugin control")
 
 plugin_group.add_option(
-    "--plugin-enable",
     "--enable-plugin",
+    "--plugin-enable",
     action="append",
     dest="plugins_enabled",
+    metavar="PLUGIN_NAME",
     default=[],
     help="""\
 Enabled plugins. Must be plug-in names. Use --plugin-list to query the
@@ -998,10 +1231,11 @@ full list and exit. Default empty.""",
 )
 
 plugin_group.add_option(
-    "--plugin-disable",
     "--disable-plugin",
+    "--plugin-disable",
     action="append",
     dest="plugins_disabled",
+    metavar="PLUGIN_NAME",
     default=[],
     help="""\
 Disabled plugins. Must be plug-in names. Use --plugin-list to query the
@@ -1015,9 +1249,9 @@ plugin_group.add_option(
     default=True,
     help="""\
 Plugins can detect if they might be used, and the you can disable the warning
-via --plugin-disable=plugin-that-warned, or you can use this option to disable
+via "--disable-plugin=plugin-that-warned", or you can use this option to disable
 the mechanism entirely, which also speeds up compilation slightly of course as
-this detection code is run in vain once you are certain of which plug-ins to
+this detection code is run in vain once you are certain of which plugins to
 use. Defaults to off.""",
 )
 
@@ -1037,13 +1271,26 @@ plugin_group.add_option(
     "--user-plugin",
     action="append",
     dest="user_plugins",
+    metavar="PATH",
     default=[],
     help="The file name of user plugin. Can be given multiple times. Default empty.",
 )
 
+plugin_group.add_option(
+    "--persist-source-changes",
+    action="store_true",
+    dest="persist_source_changes",
+    default=False,
+    help="""\
+Write source changes to original Python files. Use with care. May need
+permissions, best for use in a virtualenv to debug if plugin code
+changes work with standard Python or to benefit from bloat removal
+even with pure Python. Default False.""",
+)
+
 
 def _considerPluginOptions(logger):
-    # Recursive dependency on plugins during parsing of command line.
+    # Cyclic dependency on plugins during parsing of command line.
     from nuitka.plugins.Plugins import (
         addPluginCommandLineOptions,
         addUserPluginCommandLineOptions,
@@ -1054,12 +1301,14 @@ def _considerPluginOptions(logger):
             plugin_names = arg[16:]
             if "=" in plugin_names:
                 logger.sysexit(
-                    "Error, plugin options format changed. Use '--plugin-enable=%s --help' to know new options."
+                    "Error, plugin options format changed. Use '--enable-plugin=%s --help' to know new options."
                     % plugin_names.split("=", 1)[0]
                 )
 
             addPluginCommandLineOptions(
-                parser=parser, plugin_names=plugin_names.split(",")
+                parser=parser,
+                plugin_names=plugin_names.split(","),
+                data_files_tags=data_files_tags,
             )
 
         if arg.startswith("--user-plugin="):
@@ -1070,7 +1319,9 @@ def _considerPluginOptions(logger):
                     % plugin_name.split("=", 1)[0]
                 )
 
-            addUserPluginCommandLineOptions(parser=parser, filename=plugin_name)
+            addUserPluginCommandLineOptions(
+                parser=parser, filename=plugin_name, data_files_tags=data_files_tags
+            )
 
 
 def _expandProjectArg(arg, filename_arg, for_eval):
@@ -1081,11 +1332,22 @@ def _expandProjectArg(arg, filename_arg, for_eval):
             return value
 
     values = {
-        "OS": wrap(Utils.getOS()),
-        "Arch": wrap(Utils.getArchitecture()),
+        "OS": wrap(getOS()),
+        "Arch": wrap(getArchitecture()),
+        "Flavor": wrap(_getPythonFlavor()),
         "Version": getNuitkaVersion(),
+        "Commercial": wrap(getCommercialVersion()),
         "MAIN_DIRECTORY": wrap(os.path.dirname(filename_arg) or "."),
     }
+
+    if getOS() != "Linux":
+        dist_info = "N/A", "0"
+    else:
+        dist_info = getLinuxDistribution()
+
+    values["Linux_Distribution_Name"] = dist_info[0]
+    values["Linux_Distribution_Version"] = dist_info[1]
+
     arg = arg.format(**values)
 
     return arg
@@ -1171,6 +1433,9 @@ def _getProjectOptions(logger, filename_arg, module_mode):
             elif command == "":
                 arg = re.sub(r"""^([\w-]*=)(['"])(.*)\2$""", r"\1\3", arg.lstrip())
 
+                if not arg:
+                    continue
+
                 yield _expandProjectArg(arg, filename_arg, for_eval=False)
             else:
                 assert False, (command, line)
