@@ -19,14 +19,11 @@
 
 """
 
+import sys
+from abc import abstractmethod
+
 from nuitka import Options
-from nuitka.__past__ import (  # pylint: disable=I0021,redefined-builtin
-    GenericAlias,
-    iterItems,
-    long,
-    unicode,
-    xrange,
-)
+from nuitka.__past__ import GenericAlias, iterItems, long, unicode, xrange
 from nuitka.Builtins import (
     builtin_anon_values,
     builtin_exception_values_list,
@@ -37,11 +34,35 @@ from nuitka.Constants import (
     isConstant,
     isHashable,
     isMutable,
+    the_empty_dict,
+    the_empty_frozenset,
+    the_empty_list,
+    the_empty_set,
+    the_empty_tuple,
 )
 from nuitka.PythonVersions import python_version
 from nuitka.Tracing import optimization_logger
 
 from .ExpressionBases import CompileTimeConstantExpressionBase
+from .ExpressionShapeMixins import (
+    ExpressionBoolShapeExactMixin,
+    ExpressionBytearrayShapeExactMixin,
+    ExpressionBytesShapeExactMixin,
+    ExpressionComplexShapeExactMixin,
+    ExpressionDictShapeExactMixin,
+    ExpressionEllipsisShapeExactMixin,
+    ExpressionFloatShapeExactMixin,
+    ExpressionFrozensetShapeExactMixin,
+    ExpressionIntShapeExactMixin,
+    ExpressionListShapeExactMixin,
+    ExpressionLongShapeExactMixin,
+    ExpressionNoneShapeExactMixin,
+    ExpressionSetShapeExactMixin,
+    ExpressionSliceShapeExactMixin,
+    ExpressionStrShapeExactMixin,
+    ExpressionTupleShapeExactMixin,
+    ExpressionUnicodeShapeExactMixin,
+)
 from .IterationHandles import (
     ConstantBytearrayIterationHandle,
     ConstantBytesIterationHandle,
@@ -59,24 +80,8 @@ from .NodeMakingHelpers import (
     wrapExpressionWithSideEffects,
 )
 from .shapes.BuiltinTypeShapes import (
-    tshape_bool,
-    tshape_bytearray,
-    tshape_bytes,
-    tshape_complex,
-    tshape_dict,
-    tshape_ellipsis,
-    tshape_float,
-    tshape_frozenset,
-    tshape_int,
-    tshape_list,
-    tshape_long,
-    tshape_none,
-    tshape_set,
-    tshape_slice,
-    tshape_str,
-    tshape_tuple,
+    tshape_namedtuple,
     tshape_type,
-    tshape_unicode,
     tshape_xrange,
 )
 
@@ -120,6 +125,8 @@ class ExpressionConstantUntrackedRefBase(CompileTimeConstantExpressionBase):
         return self, None, None
 
     def computeExpressionCall(self, call_node, call_args, call_kw, trace_collection):
+        trace_collection.onExceptionRaiseExit(TypeError)
+
         # The arguments don't matter. All constant values cannot be called, and
         # we just need to make and error out of that.
         new_node = wrapExpressionWithSideEffects(
@@ -133,17 +140,19 @@ class ExpressionConstantUntrackedRefBase(CompileTimeConstantExpressionBase):
             side_effects=call_node.extractSideEffectsPreCall(),
         )
 
-        trace_collection.onExceptionRaiseExit(TypeError)
-
         return (
             new_node,
             "new_raise",
-            "Predicted call of constant value to exception raise.",
+            "Predicted call of constant %s value to exception raise."
+            % type(self.constant),
         )
 
     def getCompileTimeConstant(self):
         return self.constant
 
+    def getComparisonValue(self):
+        return self.constant
+
     @staticmethod
     def getIterationHandle():
         return None
@@ -156,11 +165,11 @@ class ExpressionConstantUntrackedRefBase(CompileTimeConstantExpressionBase):
         # This is expected to be overloaded by child classes.
         assert False, self
 
-    def extractUnhashableNode(self):
+    def extractUnhashableNodeType(self):
         value = getUnhashableConstant(self.constant)
 
         if value is not None:
-            return makeConstantRefNode(constant=value, source_ref=self.source_ref)
+            return makeConstantRefNode(constant=type(value), source_ref=self.source_ref)
 
     @staticmethod
     def isNumberConstant():
@@ -219,9 +228,11 @@ class ExpressionConstantUntrackedRefBase(CompileTimeConstantExpressionBase):
         else:
             return None
 
+    @abstractmethod
     def isIterableConstant(self):
-        # This is expected to be overloaded by child classes.
-        assert False, self
+        """Is the constant type iterable."""
+        # This is expected to be overloaded by child classes, but it's actually wasteful
+        # to use it, we should have overloads of using methods too.
 
     def getIterationLength(self):
         # This is expected to be overloaded by child classes if they are iterable
@@ -252,6 +263,8 @@ class ExpressionConstantRefBase(ExpressionConstantUntrackedRefBase):
     large lists are from computation or from user literals.
     """
 
+    # Base classes can be abstract, pylint: disable=I0021,abstract-method
+
     __slots__ = ("user_provided",)
 
     def __init__(self, constant, user_provided, source_ref):
@@ -305,7 +318,9 @@ class ExpressionConstantRefBase(ExpressionConstantUntrackedRefBase):
             return None
 
 
-class ExpressionConstantNoneRef(ExpressionConstantUntrackedRefBase):
+class ExpressionConstantNoneRef(
+    ExpressionNoneShapeExactMixin, ExpressionConstantUntrackedRefBase
+):
     kind = "EXPRESSION_CONSTANT_NONE_REF"
 
     __slots__ = ()
@@ -319,24 +334,22 @@ class ExpressionConstantNoneRef(ExpressionConstantUntrackedRefBase):
     def getDetails():
         return {}
 
-    @staticmethod
-    def getTypeShape():
-        return tshape_none
-
     @staticmethod
     def isMutable():
         return False
 
-    @staticmethod
-    def isKnownToBeHashable():
-        return True
-
     @staticmethod
     def isIterableConstant():
         return False
 
 
-class ExpressionConstantBoolRefBase(ExpressionConstantUntrackedRefBase):
+class ExpressionConstantBoolRefBase(
+    ExpressionBoolShapeExactMixin, ExpressionConstantUntrackedRefBase
+):
+    @staticmethod
+    def isExpressionConstantBoolRef():
+        return True
+
     def computeExpressionBool(self, trace_collection):
         # Best case already.
         pass
@@ -345,10 +358,6 @@ class ExpressionConstantBoolRefBase(ExpressionConstantUntrackedRefBase):
     def getDetails():
         return {}
 
-    @staticmethod
-    def getTypeShape():
-        return tshape_bool
-
     @staticmethod
     def isMutable():
         return False
@@ -412,7 +421,9 @@ class ExpressionConstantFalseRef(ExpressionConstantBoolRefBase):
         return 0
 
 
-class ExpressionConstantEllipsisRef(ExpressionConstantUntrackedRefBase):
+class ExpressionConstantEllipsisRef(
+    ExpressionEllipsisShapeExactMixin, ExpressionConstantUntrackedRefBase
+):
     kind = "EXPRESSION_CONSTANT_ELLIPSIS_REF"
 
     __slots__ = ()
@@ -426,30 +437,18 @@ class ExpressionConstantEllipsisRef(ExpressionConstantUntrackedRefBase):
     def getDetails():
         return {}
 
-    @staticmethod
-    def getTypeShape():
-        return tshape_ellipsis
-
     @staticmethod
     def isMutable():
         return False
 
-    @staticmethod
-    def isKnownToBeHashable():
-        return True
-
     @staticmethod
     def isIterableConstant():
         return False
 
-    @staticmethod
-    def getTruthValue():
-        """Return known truth value."""
 
-        return True
-
-
-class ExpressionConstantDictRef(ExpressionConstantRefBase):
+class ExpressionConstantDictRef(
+    ExpressionDictShapeExactMixin, ExpressionConstantRefBase
+):
     kind = "EXPRESSION_CONSTANT_DICT_REF"
 
     def __init__(self, constant, user_provided, source_ref):
@@ -461,14 +460,6 @@ class ExpressionConstantDictRef(ExpressionConstantRefBase):
     def isExpressionConstantDictRef():
         return True
 
-    @staticmethod
-    def getTypeShape():
-        return tshape_dict
-
-    @staticmethod
-    def hasShapeDictionaryExact():
-        return True
-
     @staticmethod
     def isMutable():
         return True
@@ -511,7 +502,14 @@ class ExpressionConstantDictRef(ExpressionConstantRefBase):
 
         for key, value in iterItems(self.constant):
             pairs.append(
-                (key, makeConstantRefNode(constant=value, source_ref=self.source_ref))
+                (
+                    key,
+                    makeConstantRefNode(
+                        constant=value,
+                        user_provided=self.user_provided,
+                        source_ref=self.source_ref,
+                    ),
+                )
             )
 
         return pairs
@@ -546,9 +544,6 @@ class EmptyContainerMixin(object):
         return False
 
 
-_the_empty_dict = {}
-
-
 class ExpressionConstantDictEmptyRef(EmptyContainerMixin, ExpressionConstantDictRef):
     kind = "EXPRESSION_CONSTANT_DICT_EMPTY_REF"
 
@@ -557,13 +552,15 @@ class ExpressionConstantDictEmptyRef(EmptyContainerMixin, ExpressionConstantDict
     def __init__(self, user_provided, source_ref):
         ExpressionConstantDictRef.__init__(
             self,
-            constant=_the_empty_dict,
+            constant=the_empty_dict,
             user_provided=user_provided,
             source_ref=source_ref,
         )
 
 
-class ExpressionConstantTupleRef(ExpressionConstantRefBase):
+class ExpressionConstantTupleRef(
+    ExpressionTupleShapeExactMixin, ExpressionConstantRefBase
+):
     kind = "EXPRESSION_CONSTANT_TUPLE_REF"
 
     __slots__ = ()
@@ -577,10 +574,6 @@ class ExpressionConstantTupleRef(ExpressionConstantRefBase):
     def isExpressionConstantTupleRef():
         return True
 
-    @staticmethod
-    def getTypeShape():
-        return tshape_tuple
-
     @staticmethod
     def isMutable():
         return False
@@ -627,9 +620,6 @@ class ExpressionConstantTupleMutableRef(ExpressionConstantTupleRef):
         return False
 
 
-_the_empty_tuple = ()
-
-
 class ExpressionConstantTupleEmptyRef(EmptyContainerMixin, ExpressionConstantTupleRef):
     kind = "EXPRESSION_CONSTANT_TUPLE_EMPTY_REF"
 
@@ -638,13 +628,15 @@ class ExpressionConstantTupleEmptyRef(EmptyContainerMixin, ExpressionConstantTup
     def __init__(self, user_provided, source_ref):
         ExpressionConstantTupleRef.__init__(
             self,
-            constant=_the_empty_tuple,
+            constant=the_empty_tuple,
             user_provided=user_provided,
             source_ref=source_ref,
         )
 
 
-class ExpressionConstantListRef(ExpressionConstantRefBase):
+class ExpressionConstantListRef(
+    ExpressionListShapeExactMixin, ExpressionConstantRefBase
+):
     kind = "EXPRESSION_CONSTANT_LIST_REF"
 
     __slots__ = ()
@@ -658,18 +650,10 @@ class ExpressionConstantListRef(ExpressionConstantRefBase):
     def isExpressionConstantListRef():
         return True
 
-    @staticmethod
-    def getTypeShape():
-        return tshape_list
-
     @staticmethod
     def isMutable():
         return True
 
-    @staticmethod
-    def isKnownToBeHashable():
-        return False
-
     @staticmethod
     def isIterableConstant():
         return True
@@ -697,9 +681,6 @@ class ExpressionConstantListRef(ExpressionConstantRefBase):
         )
 
 
-_the_empty_list = []
-
-
 class ExpressionConstantListEmptyRef(EmptyContainerMixin, ExpressionConstantListRef):
     kind = "EXPRESSION_CONSTANT_LIST_EMPTY_REF"
 
@@ -708,13 +689,13 @@ class ExpressionConstantListEmptyRef(EmptyContainerMixin, ExpressionConstantList
     def __init__(self, user_provided, source_ref):
         ExpressionConstantListRef.__init__(
             self,
-            constant=_the_empty_list,
+            constant=the_empty_list,
             user_provided=user_provided,
             source_ref=source_ref,
         )
 
 
-class ExpressionConstantSetRef(ExpressionConstantRefBase):
+class ExpressionConstantSetRef(ExpressionSetShapeExactMixin, ExpressionConstantRefBase):
     kind = "EXPRESSION_CONSTANT_SET_REF"
 
     __slots__ = ()
@@ -728,10 +709,6 @@ class ExpressionConstantSetRef(ExpressionConstantRefBase):
     def isExpressionConstantSetRef():
         return True
 
-    @staticmethod
-    def getTypeShape():
-        return tshape_set
-
     @staticmethod
     def isMutable():
         return True
@@ -767,9 +744,6 @@ class ExpressionConstantSetRef(ExpressionConstantRefBase):
         )
 
 
-_the_empty_set = set()
-
-
 class ExpressionConstantSetEmptyRef(EmptyContainerMixin, ExpressionConstantSetRef):
     kind = "EXPRESSION_CONSTANT_SET_EMPTY_REF"
 
@@ -778,13 +752,15 @@ class ExpressionConstantSetEmptyRef(EmptyContainerMixin, ExpressionConstantSetRe
     def __init__(self, user_provided, source_ref):
         ExpressionConstantSetRef.__init__(
             self,
-            constant=_the_empty_set,
+            constant=the_empty_set,
             user_provided=user_provided,
             source_ref=source_ref,
         )
 
 
-class ExpressionConstantFrozensetRef(ExpressionConstantRefBase):
+class ExpressionConstantFrozensetRef(
+    ExpressionFrozensetShapeExactMixin, ExpressionConstantRefBase
+):
     kind = "EXPRESSION_CONSTANT_FROZENSET_REF"
 
     __slots__ = ()
@@ -798,18 +774,10 @@ class ExpressionConstantFrozensetRef(ExpressionConstantRefBase):
     def isExpressionConstantFrozensetRef():
         return True
 
-    @staticmethod
-    def getTypeShape():
-        return tshape_frozenset
-
     @staticmethod
     def isMutable():
         return False
 
-    @staticmethod
-    def isKnownToBeHashable():
-        return True
-
     @staticmethod
     def isIterableConstant():
         return True
@@ -837,9 +805,6 @@ class ExpressionConstantFrozensetRef(ExpressionConstantRefBase):
         )
 
 
-_the_empty_frozenset = frozenset()
-
-
 class ExpressionConstantFrozensetEmptyRef(
     EmptyContainerMixin, ExpressionConstantFrozensetRef
 ):
@@ -850,13 +815,15 @@ class ExpressionConstantFrozensetEmptyRef(
     def __init__(self, user_provided, source_ref):
         ExpressionConstantFrozensetRef.__init__(
             self,
-            constant=_the_empty_frozenset,
+            constant=the_empty_frozenset,
             user_provided=user_provided,
             source_ref=source_ref,
         )
 
 
-class ExpressionConstantIntRef(ExpressionConstantUntrackedRefBase):
+class ExpressionConstantIntRef(
+    ExpressionIntShapeExactMixin, ExpressionConstantUntrackedRefBase
+):
     kind = "EXPRESSION_CONSTANT_INT_REF"
 
     __slots__ = ()
@@ -870,18 +837,10 @@ class ExpressionConstantIntRef(ExpressionConstantUntrackedRefBase):
     def isExpressionConstantIntRef():
         return True
 
-    @staticmethod
-    def getTypeShape():
-        return tshape_int
-
     @staticmethod
     def isMutable():
         return False
 
-    @staticmethod
-    def isKnownToBeHashable():
-        return True
-
     @staticmethod
     def isNumberConstant():
         return True
@@ -898,7 +857,9 @@ class ExpressionConstantIntRef(ExpressionConstantUntrackedRefBase):
         return False
 
 
-class ExpressionConstantLongRef(ExpressionConstantRefBase):
+class ExpressionConstantLongRef(
+    ExpressionLongShapeExactMixin, ExpressionConstantRefBase
+):
     kind = "EXPRESSION_CONSTANT_LONG_REF"
 
     __slots__ = ()
@@ -912,18 +873,10 @@ class ExpressionConstantLongRef(ExpressionConstantRefBase):
     def isExpressionConstantLongRef():
         return True
 
-    @staticmethod
-    def getTypeShape():
-        return tshape_long
-
     @staticmethod
     def isMutable():
         return False
 
-    @staticmethod
-    def isKnownToBeHashable():
-        return True
-
     @staticmethod
     def isNumberConstant():
         return True
@@ -942,7 +895,7 @@ class ExpressionConstantLongRef(ExpressionConstantRefBase):
         return False
 
 
-class ExpressionConstantStrRef(ExpressionConstantRefBase):
+class ExpressionConstantStrRef(ExpressionStrShapeExactMixin, ExpressionConstantRefBase):
     kind = "EXPRESSION_CONSTANT_STR_REF"
 
     __slots__ = ()
@@ -956,18 +909,10 @@ class ExpressionConstantStrRef(ExpressionConstantRefBase):
     def isExpressionConstantStrRef():
         return True
 
-    @staticmethod
-    def getTypeShape():
-        return tshape_str
-
     @staticmethod
     def isMutable():
         return False
 
-    @staticmethod
-    def isKnownToBeHashable():
-        return True
-
     @staticmethod
     def isIterableConstant():
         return True
@@ -1003,7 +948,9 @@ class ExpressionConstantStrEmptyRef(EmptyContainerMixin, ExpressionConstantStrRe
         )
 
 
-class ExpressionConstantUnicodeRef(ExpressionConstantRefBase):
+class ExpressionConstantUnicodeRef(
+    ExpressionUnicodeShapeExactMixin, ExpressionConstantRefBase
+):
     kind = "EXPRESSION_CONSTANT_UNICODE_REF"
 
     __slots__ = ()
@@ -1017,18 +964,10 @@ class ExpressionConstantUnicodeRef(ExpressionConstantRefBase):
     def isExpressionConstantUnicodeRef():
         return True
 
-    @staticmethod
-    def getTypeShape():
-        return tshape_unicode
-
     @staticmethod
     def isMutable():
         return False
 
-    @staticmethod
-    def isKnownToBeHashable():
-        return True
-
     @staticmethod
     def isIterableConstant():
         return True
@@ -1060,7 +999,9 @@ class ExpressionConstantUnicodeEmptyRef(
         )
 
 
-class ExpressionConstantBytesRef(ExpressionConstantRefBase):
+class ExpressionConstantBytesRef(
+    ExpressionBytesShapeExactMixin, ExpressionConstantRefBase
+):
     kind = "EXPRESSION_CONSTANT_BYTES_REF"
 
     def __init__(self, constant, user_provided, source_ref):
@@ -1072,18 +1013,10 @@ class ExpressionConstantBytesRef(ExpressionConstantRefBase):
     def isExpressionConstantBytesRef():
         return True
 
-    @staticmethod
-    def getTypeShape():
-        return tshape_bytes
-
     @staticmethod
     def isMutable():
         return False
 
-    @staticmethod
-    def isKnownToBeHashable():
-        return True
-
     @staticmethod
     def isIterableConstant():
         return True
@@ -1113,7 +1046,9 @@ class ExpressionConstantBytesEmptyRef(EmptyContainerMixin, ExpressionConstantByt
         )
 
 
-class ExpressionConstantBytearrayRef(ExpressionConstantRefBase):
+class ExpressionConstantBytearrayRef(
+    ExpressionBytearrayShapeExactMixin, ExpressionConstantRefBase
+):
     kind = "EXPRESSION_CONSTANT_BYTEARRAY_REF"
 
     def __init__(self, constant, user_provided, source_ref):
@@ -1125,18 +1060,10 @@ class ExpressionConstantBytearrayRef(ExpressionConstantRefBase):
     def isExpressionConstantBytearrayRef():
         return True
 
-    @staticmethod
-    def getTypeShape():
-        return tshape_bytearray
-
     @staticmethod
     def isMutable():
         return True
 
-    @staticmethod
-    def isKnownToBeHashable():
-        return False
-
     @staticmethod
     def isIterableConstant():
         return True
@@ -1164,7 +1091,9 @@ class ExpressionConstantBytearrayRef(ExpressionConstantRefBase):
         )
 
 
-class ExpressionConstantFloatRef(ExpressionConstantUntrackedRefBase):
+class ExpressionConstantFloatRef(
+    ExpressionFloatShapeExactMixin, ExpressionConstantUntrackedRefBase
+):
     kind = "EXPRESSION_CONSTANT_FLOAT_REF"
 
     __slots__ = ()
@@ -1178,18 +1107,10 @@ class ExpressionConstantFloatRef(ExpressionConstantUntrackedRefBase):
     def isExpressionConstantFloatRef():
         return True
 
-    @staticmethod
-    def getTypeShape():
-        return tshape_float
-
     @staticmethod
     def isMutable():
         return False
 
-    @staticmethod
-    def isKnownToBeHashable():
-        return True
-
     @staticmethod
     def isNumberConstant():
         return True
@@ -1199,7 +1120,9 @@ class ExpressionConstantFloatRef(ExpressionConstantUntrackedRefBase):
         return False
 
 
-class ExpressionConstantComplexRef(ExpressionConstantUntrackedRefBase):
+class ExpressionConstantComplexRef(
+    ExpressionComplexShapeExactMixin, ExpressionConstantUntrackedRefBase
+):
     kind = "EXPRESSION_CONSTANT_COMPLEX_REF"
 
     __slots__ = ()
@@ -1213,24 +1136,23 @@ class ExpressionConstantComplexRef(ExpressionConstantUntrackedRefBase):
     def isExpressionConstantComplexRef():
         return True
 
-    @staticmethod
-    def getTypeShape():
-        return tshape_complex
-
     @staticmethod
     def isMutable():
         return False
 
-    @staticmethod
-    def isKnownToBeHashable():
-        return True
-
     @staticmethod
     def isIterableConstant():
         return False
 
+    # Overload what ExpressionComplexShapeExactMixin says, for a given instance we know all.
+    @staticmethod
+    def isKnownToHaveAttribute(attribute_name):
+        return hasattr(0j, attribute_name)
 
-class ExpressionConstantSliceRef(ExpressionConstantUntrackedRefBase):
+
+class ExpressionConstantSliceRef(
+    ExpressionSliceShapeExactMixin, ExpressionConstantUntrackedRefBase
+):
     kind = "EXPRESSION_CONSTANT_SLICE_REF"
 
     __slots__ = ()
@@ -1244,18 +1166,10 @@ class ExpressionConstantSliceRef(ExpressionConstantUntrackedRefBase):
     def isExpressionConstantSliceRef():
         return True
 
-    @staticmethod
-    def getTypeShape():
-        return tshape_slice
-
     @staticmethod
     def isMutable():
         return False
 
-    @staticmethod
-    def isKnownToBeHashable():
-        return False
-
     @staticmethod
     def isIterableConstant():
         return False
@@ -1572,7 +1486,59 @@ def makeConstantRefNode(constant, source_ref, user_provided=False):
             return ExpressionConstantGenericAlias(
                 generic_alias=constant, source_ref=source_ref
             )
-
+        elif constant is sys.version_info:
+            return ExpressionConstantSysVersionInfoRef(source_ref=source_ref)
         else:
             # Missing constant type, ought to not happen, please report.
             assert False, (constant, constant_type)
+
+
+class ExpressionConstantSysVersionInfoRef(ExpressionConstantUntrackedRefBase):
+    kind = "EXPRESSION_CONSTANT_SYS_VERSION_INFO_REF"
+
+    __slots__ = ()
+
+    def __init__(self, source_ref):
+        ExpressionConstantUntrackedRefBase.__init__(
+            self, constant=sys.version_info, source_ref=source_ref
+        )
+
+    @staticmethod
+    def getDetails():
+        return {}
+
+    @staticmethod
+    def getTypeShape():
+        return tshape_namedtuple
+
+    @staticmethod
+    def isMutable():
+        return False
+
+    @staticmethod
+    def isKnownToBeHashable():
+        return True
+
+    @staticmethod
+    def isIterableConstant():
+        return True
+
+    def getIterationHandle(self):
+        return ConstantTupleIterationHandle(self)
+
+    def getIterationLength(self):
+        return len(self.constant)
+
+    def computeExpressionIter1(self, iter_node, trace_collection):
+        # For iteration, we are just a normal tuple.
+        return (
+            ExpressionConstantTupleRef(
+                self.constant, user_provided=True, source_ref=self.source_ref
+            ),
+            "new_constant",
+            """Iteration over constant sys.version_info lowered to tuple.""",
+        )
+
+    @staticmethod
+    def getTruthValue():
+        return True
